/**************************************************************************\
|
|  Test program for freestart collision for 76-step SHA-1
|
|  Pierre Karpman, NTU Singapore, Inria France
|  Thomas Peyrin, NTU Singapore
|  Marc Stevens, CWI Netherlands
|  Copyright (C) 2014 CWI, NTU, Inria
|    
|    Contact:
|    Marc Stevens
|    Cryptology Group
|    Centrum Wiskunde & Informatica
|    P.O. Box 94079, 1090 GB Amsterdam, Netherlands
|    marc@marc-stevens.nl
|
|  Permission is hereby granted, free of charge, to any person obtaining a copy
|  of this software and associated documentation files (the "Software"), to deal
|  in the Software without restriction, including without limitation the rights
|  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
|  copies of the Software, and to permit persons to whom the Software is
|  furnished to do so, subject to the following conditions:
| 
|  The above copyright notice and this permission notice shall be included in
|  all copies or substantial portions of the Software.
| 
|  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
|  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
|  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
|  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
|  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
|  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
|  THE SOFTWARE.
|
\**************************************************************************/

#include <stdio.h>
#include <stdint.h>
#include <assert.h>
#include <iostream>
#include <iomanip>


#define K0 0x5a827999
#define K1 0x6ed9eba1
#define K2 0x8f1bbcdc
#define K3 0xca62c1d6


inline uint32_t rotate_left(uint32_t x, uint32_t r)
{
	return ((x << r) | (x >> (32 - r)));
}

inline uint32_t sha1_f1(uint32_t x, uint32_t y, uint32_t z)
{
	return ((x & y) | (~x & z));
}

inline uint32_t sha1_f2(uint32_t x, uint32_t y, uint32_t z)
{
	return (x ^ y ^ z);
}

inline uint32_t sha1_f3(uint32_t x, uint32_t y, uint32_t z)
{
	return ((x & y) | (x & z) | (y & z));
}

inline uint32_t sha1_f4(uint32_t x, uint32_t y, uint32_t z)
{
	return (x ^ y ^ z);
}

inline void sha1_step_round1(int t, uint32_t Q[85], uint32_t m[80])
{
	Q[4+t+1] = rotate_left(Q[4+t],5) + rotate_left(Q[4+t-4],30) + sha1_f1(Q[4+t-1],rotate_left(Q[4+t-2],30),rotate_left(Q[4+t-3],30)) + K0 + m[t];
}
inline void sha1_step_round2(int t, uint32_t Q[85], uint32_t m[80])
{
	Q[4+t+1] = rotate_left(Q[4+t],5) + rotate_left(Q[4+t-4],30) + sha1_f2(Q[4+t-1],rotate_left(Q[4+t-2],30),rotate_left(Q[4+t-3],30)) + K1 + m[t];
}
inline void sha1_step_round3(int t, uint32_t Q[85], uint32_t m[80])
{
	Q[4+t+1] = rotate_left(Q[4+t],5) + rotate_left(Q[4+t-4],30) + sha1_f3(Q[4+t-1],rotate_left(Q[4+t-2],30),rotate_left(Q[4+t-3],30)) + K2 + m[t];
}
inline void sha1_step_round4(int t, uint32_t Q[85], uint32_t m[80])
{
	Q[4+t+1] = rotate_left(Q[4+t],5) + rotate_left(Q[4+t-4],30) + sha1_f4(Q[4+t-1],rotate_left(Q[4+t-2],30),rotate_left(Q[4+t-3],30)) + K3 + m[t];
}

void sha1_me(uint32_t M[80])
{
	for (int i = 16; i < 80; ++i)
	{
		M[i] = rotate_left(M[i - 3] ^ M[i - 8] ^ M[i - 14] ^ M[i - 16], 1);
	}
}

void sha1_76(uint32_t out_cv[5], uint32_t Q[85], const uint32_t in_cv[5], const uint32_t m[16])
{
	uint32_t W[80];
	for (unsigned i = 0; i < 16; ++i)
		W[i] = m[i];
	sha1_me(W);
	
	Q[4] = in_cv[0];
	Q[3] = in_cv[1];
	Q[2] = rotate_left(in_cv[2],2);
	Q[1] = rotate_left(in_cv[3],2);
	Q[0] = rotate_left(in_cv[4],2);
	
	for (int t =  0; t < 20; ++t)
		sha1_step_round1(t, Q, W);
	for (int t = 20; t < 40; ++t)
		sha1_step_round2(t, Q, W);
	for (int t = 40; t < 60; ++t)
		sha1_step_round3(t, Q, W);
	for (int t = 60; t < 76; ++t)
		sha1_step_round4(t, Q, W);
		
	out_cv[0] = in_cv[0] + Q[4+76];
	out_cv[1] = in_cv[1] + Q[4+75];
	out_cv[2] = in_cv[2] + rotate_left(Q[4+74],30);
	out_cv[3] = in_cv[3] + rotate_left(Q[4+73],30);
	out_cv[4] = in_cv[4] + rotate_left(Q[4+72],30);
}

void test()
{
	uint32_t iv1[5] = { 0x81bf2306, 0x41b83b5c, 0x03e9a78f, 0xba5028d5, 0xfc508788 };
	uint32_t iv2[5] = { 0x81bf2306, 0x41b83b5d, 0x83e9a78f, 0xba5028d5, 0xfc508788 };

	uint32_t m1[16] = { 0x46fa5a88, 0xf4f0c7f0, 0xb8dedbec, 0x951e2588,
	                    0x7734fdf5, 0x4c42c497, 0x52d7d8f9, 0x5f1452ea,
	                    0xb49e93b2, 0x91c23071, 0xc70f359b, 0x8abacfaf, 
	                    0xb37ffb27, 0x3dfe7fad, 0x7ade5695, 0x20fd7cea  };
	uint32_t m2[16] = { 0x46fa5a98, 0xf0f0c7ec, 0x74dedbf8, 0x991e258a, 
	                    0xb734fde5, 0xf842c48b, 0x6ed7d8fd, 0xe31452f0, 
	                    0x949e93a2, 0xb5c2306d, 0x2b0f358f, 0x86bacfad, 
	                    0x737ffb37, 0x89fe7fb1, 0x56de5691, 0x9cfd7cf2  };

	uint32_t cv1[5], cv2[5];
	uint32_t Q1[85], Q2[85];
        
	sha1_76(cv1, Q1, iv1, m1);
	sha1_76(cv2, Q2, iv2, m2);

	for (int t = -4; t <= 76; ++t)
	{
	    std::cout << std::setw(2) << t << " ";
	    uint32_t q1 = Q1[4+t], q2 = Q2[4+t];
	    for (int b = 31; b >= 0; --b)
            {
                uint32_t q1b = (q1>>b)&1, q2b = (q2>>b)&1;
                if (q1b == q2b) {
                    std::cout << q1b;
                } else if (q1b == 0) {
                    std::cout << "+";
                } else if (q1b == 1) {
                    std::cout << "-";
                }
            }
            std::cout << std::endl;
            if (t == 0 || t == 71 || t == 76)
                std::cout << std::endl;
	}
	for (unsigned i = 0; i < 5; ++i)
	{
	    std::cout << std::setw(2) << i << " ";
	    uint32_t q1 = cv1[i], q2 = cv2[i];
	    for (int b = 31; b >= 0; --b)
            {
                uint32_t q1b = (q1>>b)&1, q2b = (q2>>b)&1;
                if (q1b == q2b) {
                    std::cout << q1b;
                } else if (q1b == 0) {
                    std::cout << "+";
                } else if (q1b == 1) {
                    std::cout << "-";
                }
            }
            std::cout << std::endl;
	}

	std::cout << std::endl;
	for (unsigned i = 0; i < 5; ++i)
	{
		std::cout << std::setw(2) << i << " " << std::hex << cv1[i] << " " << std::hex << cv2[i] << std::dec << std::endl;
	}
}

int main(int argc, char **argv)
{
	test();
	return 0;
}
